@@ -14,7 +14,4 @@ VECTYPE calculate_norm(const MATTYPE& M);
 int my_ceil(float num);
 
 
-double find_one_lambda_cpp(const arma::vec& cluster_size, const arma::vec& range);
-
-arma::vec find_lambda_cpp(const arma::vec& cluster_size, const arma::vec& range,
-                          const std::vector<int>& B_vec);
+arma::vec find_lambda_cpp(const float alpha, const arma::vec& cluster_E);
